#
# ex:tw=0
#
# $Id: types.rb,v 1.46 2002/11/04 23:44:41 eivind Exp $
#
# Copyright (C) 2002 Eivind Eklund <eivind@FreeBSD.org>
# This code is available under the same license as Ruby.
#
# Module to easily add type checking to methods.
#
# Usage:
# Note that "is" below actually means a kind_of? relation unless
# specifically said to be otherwise.
#
# class MyClass
#	extend TypeChecker
#
#	typesig String, String		# Types for next defined method
#	def example1(string1, string2)
#		... do stuff, knowing that string1 and string2 are Strings ...
#	end
#	
#	#
#	# We also support allowing a parameter to have multiple types:
#	#
#	typesig String, [Array, Hash]
#	def example2(string, array_or_hash)
#		... here, we know that string is a String, and array_or_hash
#		is either an Array or a Hash (or a subclass thereof) ...
#	end
#	#
#	# And we can specify that the rest of a parameter list should be of a
#	# single type
#	#
#	typesig Hash, :rest, String
#	def example3(hash, *strings)
#		... Know that hash is a Hash, and strings is an array of
#		Strings ...
#	end
#	#
#	# Sometimes, it is convenient to limit a set of types by what messages
#	# they understand
#	#
#	typesig Type::Respond(:type, :name)
#	def example3(something)
#		print "Something: " + something.name + "\n"
#		if something.type == "Wally"
#			# .. Handle wallies here.
#		end
#	end
#	#
#	# For a limited set of cases, we want to be able to match an EXACT
#	# type.  If we specify with :exact, the next type will be checked with
#	# instance_of? instead of kind_of?
#	#
#	typesig String, :exact, Module
#	def example4(string, module)
#		... Here we know that string is some kind of string,
#		and we know that module is a module, and not a Class ...
#	end
#	#
#	# We also support having something that must match *all* of a
#	# particular set of types - which can be used to test for the
#	# type having mixed in a series of modules, for instance.
#	#
#	# The below looks for subclasses of Array that have been defined
#	# Comparable
#	#
#	typesig Type::All(Array, Comparable)
#	def example5(comparablearray)
#	end
#	#
#	# We support having optional parameters
#	#
#	# XXX :optional is not implemented yet
#	# XXX This screws the arity check over; need code re-org
#	# XXX The example below sucks, anyway.
#	typesig String, String, :optional, String
#	def example6(string1, string2, string3=nil)
#	end
#	#
#	# Also, we can do "argument grouping" - create a
#	# type that "consume" more than one argument, e.g.
#	# for a varargs list that consists of String, Integer
#	# pairs:
#	# XXX Type::Multi is not implemented yet
#	typesig Class, :rest, Type::Multi(String, Integer)
#	def example7(klass, *pairs)
#		... and pairs[0+2*x] will be a String, while
#		pairs[1+2*x] will be an Integer ...
#	end
#
#	#
#	# Field limitations - where we actually look at the data in the field
#	#
#	# XXX NONE of these are implemented yet
#	#
#
#	#
#	# Formatting for a method that takes printf-like arguments, with
#	# the arguments
#	#
#	typesig IO, Type::Printf
#	def examplefprintf(file, format, *rest)
#	end
#	#
#	# Limit to strings matching a particular regexp
#	# to_s will be called, so if you want to assure that it IS a string,
#	# you need to do something like
#	# typesig :all, [String, Type::Regexp(/^\d+/)]
#	#
#	typesig Type::Regexp(/something/)
#
#	XXX Verify numeric range
#	XXX Verify hash keys
#	XXX Verify array length
#	XXX Verify generic properties (by use of an in-place function)
#	XXX Go through the list of built-in classes for Ruby and see if there
#	are more of them we want to have specific types for restrictions on
#	XXX Pure NotNil
#	XXX Pure IsNil
#	XXX Perhaps some way of specifying "except" for a set?
#	XXX In order to avoid having too many methods: Some way of specifying
#	that a set of type restrictions instead of applying directly to the
#	object in question applies to the return value from a particular
#	method in the object - e.g, Type::Return(:length, Type::Range(1, nil))
#	to specify that you want an Array object to have at least one entry
#
#	XXX Look at integration with Types.rb from matju
#	XXX Look at abandoning the name Type, as it conflicts with matju
#
# end
#
#
#
#	If you see any functionality you'd like that is NOT described above,
#	please DO request it.  
#
#
#

module TypeChecker
	#
	# Avoid the possibility of using this module by include instead of
	# extend
	#
	def TypeChecker.append_features(klass)
		raise NotImplementedError, "#{name} is not intended to be included.  Use 'extend #{name}'", caller
	end
	#
	# Make other method_added available to others as well.
	#
	def TypeChecker.extend_object(klass)
		klass.module_eval {
			class <<self
				if method_defined? :method_added
					alias_method :tc_method_added, :method_added
					remove_method :method_added
				else
					def tc_method_added(unused) end
				end
			end
		}
		super(klass)
	end

	#
	# Add type checks to a live method in a class (do not defer until the object 
	#
	# Note that method_added will be called for the typechecking method being
	# interjected in the call stream (unlike when typesig is used for automatic
	# addition.)
	#
	def TypeChecker.add_typesig(myclass, method, *typesig)
		typesig = TypeChecker.typesig_normalize(*typesig)
		raise ArgumentError, "#{myclass.type.name} does not respond to #{method}" \
			unless myclass.instance_method(method)
		myclass.module_eval <<-eom
			@@tc_typesig_#{method.id} = typesig
			alias_method("__tc_#{method.id}", method)
			def #{method}(*args)
				# XXX This arity check will not work, but the real call will take care of it
				err = TypeChecker.check_types(#{myclass.instance_method(method).arity}, args, *@@tc_typesig_#{method.id})
				if err
					raise ArgumentError, err, caller
				else
					return __tc_#{method.id}(*args)
				end
			end
		eom
	end

	#
	# Add a typesig to a method in a particular class that is extended by
	# TypeChecker.
	#
	def add_typesig(method, *typesig)
		TypeChecker.add_typesig(self, method, *typesig)
	end
	#
	# Add typechecks to a method in a live object.
	#
	# Note that this will result in an anonymous class being
	# created for the object in question, and is thus slightly
	# wasteful.  However, it can be very convenient.
	#
	# For example, when I had a problem that used a Hash, and I wanted to
	# make sure that I only used Strings to index the hash (because I was
	# getting data from several sources and it could arrive to me either
	# as Strings or as Integers), I could have fixed this with
	# TypeChecker.object_add_typesig(myhash, :[], String)
	# TypeChecker.object_add_typesig(myhash, :[]=, String)
	#
	def TypeChecker.object_add_typesig(object, method, *typesig)
		# XXX A temporary critical section and global variable might be faster?
		class <<object
			Thread.current["typechecker_object_add_typesig_class"] = self
		end
		add_typesig(Thread.current["typechecker_object_add_typesig_class"], method, *typesig)
	end

	#
	# Do type checks iff we have registered a set of types for this method
	#
	def method_added(id)
		return if @tc_adding_method
		if @tc_typesig_data
			# Avoid problems with self callbacks
			@tc_adding_method = true
			typesig = @tc_typesig_data
			@tc_typesig_data = nil
			TypeChecker.add_typesig(self, id, *typesig)
			@tc_adding_method = false
		end
		tc_method_added(id)
	end
	#
	# Next argument in typesig
	#
	def TypeChecker.next_arg(args, typeoffset)
		while typeoffset < args.length && args[typeoffset].kind_of?(Symbol)
			typeoffset += 1
		end
		return nil unless typeoffset < args.length
		typeoffset
	end
	private_class_method :next_arg

	#
	# Check and normalize a typesig before use.
	#
	# This verifies that all convetions are OK, so a set of types can
	# be shown as faulty at initial definition instead of first use.
	#
	def TypeChecker.typesig_normalize(*args)
		args = Type.normalize_arg_refs(args)
		args.each_index { |typeoffset|
			mytype = args[typeoffset]
			if mytype.kind_of? Module
			elsif mytype.kind_of? Array
				args[typeoffset] = Type::Any(*mytype)
				next
				mytype.each_index { |arrayoffset|
					realtype = mytype[arrayoffset]
					if realtype.kind_of? Type
					elsif realtype.kind_of? Module
					else
						raise ArgumentError, \
							"Bad type spec in array[#{arrayoffset}] at arg #{typeoffset}"
					end
				}
			elsif mytype.kind_of? Type
			elsif mytype.kind_of? Symbol
				raise ArgumentError, "#{mytype.inspect} without next parameter", caller(3) \
					if typeoffset+1 == args.length
				case mytype
				when :rest
					next_type = next_arg(args, typeoffset)
					raise ArgumentError, "No type after :rest" unless next_type
					raise ArgumentError, "More than one parameter after :rest", caller(3) \
						unless next_type+1 >= args.length
				when :exact
				else
					raise ArgumentError, "Unknown modifier: #{mytype.inspect}"
				end
			else
				raise "Invalid typesig"
			end
		}
		args
	end
	#
	# Register types for next method
	#
	def typesig(*args)
		raise "typesig called multiple times without method definition between" if @tc_typesig_data
		raise "typesig at inside method_added recursion (this should never happen)" if @tc_added_method
		@tc_typesig_data = TypeChecker.typesig_normalize(*args)
	end

	#
	# Verify the signature for a method.
	#
	def TypeChecker.check_types(arity, args, *typesig)
		if arity >= 0
			return "wrong # of arguments (#{args.length} for #{arity})" \
				unless args.length == arity
		else
			return "wrong # of arguments (#{args.length} for #{-arity - 1})" \
				unless args.length >= -arity-1
		end
		argoffset = 0
		typeoffset = 0
		argslength = args.length
		typesiglength = typesig.length
		ok = false		# Pre-initialization for Array tests; speed issue
		rest = false		# Are we processing a :rest clause?
		exact = false		# Should next argument be exact?
		while argoffset < argslength && typeoffset < typesiglength
			mytype = typesig[typeoffset]
			if mytype.kind_of? Type
				offset, string = mytype.verify(argoffset, arity, exact, args, *typesig)
				return "Arg #{argoffset} " + string if string
				argoffset += offset
			elsif mytype.kind_of? Symbol
				case mytype
				when :rest
					rest = true
					typeoffset += 1
				when :exact
					exact = 1
					typeoffset += 1 if rest
				else
					raise ArgumentError, "Unknown modifier in TypeChecker.check_types"
				end
			else
				if ((!exact && args[argoffset].kind_of?(mytype)) ||
				    (exact  && args[argoffset].instance_of?(mytype)))
					argoffset += 1
				else
					return \
"Arg #{argoffset} is of invalid type (expected #{mytype}, got #{args[argoffset].type})"
				end
			end
			unless rest
				typeoffset += 1
				exact = false if exact == 0
				exact = 0 if exact == 1
			end
		end
		return nil
	end
end


#
# Representation of types, for type checking.
# This can represent arbitarily complex types, not just classes.
# Methods in Type are generally just a short way of creating an object.
#
class Type
	def initialize(*args)
		@verify_message = "is not compatible with #{to_s}"
		@args = args
		Thread.current["Type::@verify_message"] = nil
	end
	#
	# Generate a description of the type as Ruby code
	#
	def inspect
		if @args
			return "#{type.name}(" + @args.collect { |arg| arg.inspect }.join(', ') + ")"
		else
			return "#{type.name}"
		end
	end

	# XXX It might be more proper to have other code call inspect instead
	# of to_s, but that's more complicated, and YouAintGonnaNeedIt.
	def to_s
		inspect
	end

	#
	# Verify that this particular type is OK for the arg in question.
	#
	# types:: Array of types (with this object somehow available through
	# the slot number given, though that may be through another method)
	# arity:: Method arity for the method called
	# args:: The arguments the method was called with
	# argno:: The argument number to check
	#
	# Returns:
	#	Number of arguments consumed, string for error
	#
	def verify(argno, arity, exact, args, *types)
 		if self.===(args[argno], true)
			return 1, nil
		else
			return nil, get_message()
		end
	end

	#
	# Check if this type is compatible with the argument.
	# This allows use of the different types in
	#	case object
	#	when <sometype>
	#		... do stuff ...
	#	end
	#
	# This method may call set_message to set a specific message for
	# how the type failed; see Type::Respond::=== for an example.
	#
	def ===(arg, set_message=nil)
		raise "Use of === (case/when) on type this is not valid for"
	end

	#
	# Set the message for verify.  Intended to be called by
	# === for types that do their checks there, but want to
	# customize the message based on how the object does not match.
	#
	def set_message(msg, set_message=true)
		raise "Bug in Type #{self.type}: Varying calls to set_message" \
			if Thread.current["Type::@verify_message"]
		return nil unless set_message
		Thread.current["Type::@verify_message"] = msg
		nil
	end
	private :set_message

	#
	# Get a message previously set by set_message() (or left from object
	# init)
	#
	def get_message
		msg = Thread.current["Type::@verify_message"]
		Thread.current["Type::@verify_message"] = nil
		msg = @verify_message unless msg
		msg
	end
	private :get_message

	#
	# Normalize references to args to be to the first equal arg ever passed
	# (using an internal cache).
	#
	# This assumes that == works for the arguments (as seen from the point
	# of view of typechecking) - ie, that if two objects are matched as ==,
	# either can be used.
	#
	# It also assumes that if arg.hash is implemented, the arg can be
	# placed in a hash.
	#
	# Args will be frozen unless they are modules; if you need an argument
	# to a Type that has varying behaviour based on time (usually a bad
	# idea, but may be useful for certain systems), you need to implement
	# this using an extra layer of indirection.  The use of freeze is intended
	# to avoid obscure application bugs caused by the object aliasing done
	# inside the type handling.
	#
	@@known_arg_types = {}
	def Type.normalize_ref(arg)
		arg.freeze unless arg.kind_of? Module
		argtype = arg.type
		unless @@known_arg_types[argtype.name]
			@@known_arg_types[argtype.name] = argtype.method_defined?(:hash) ? {} : []
		end
		if @@known_arg_types[argtype.name].type == Array
			index = @@known_arg_types[argtype.name].index(arg)
			if index
				return @@known_arg_types[argtype.name][index]
			else
				@@known_arg_types[argtype.name].push(arg)
				return arg
			end
		else
			# Hash
 			unless @@known_arg_types[argtype.name][arg]
				@@known_arg_types[argtype.name][arg] = arg
			end
			return @@known_arg_types[argtype.name][arg]
		end
	end

	#
	# Normalize references to args to be the first identical arg ever
	# passed in, to maximize chance of hitting the cache.
	#
	def Type.normalize_arg_refs(args)
		args.collect { |arg| normalize_ref(arg) }
	end

	#
	# For many types, the order of arguments does not matter and number of
	# repeats of an argument does not matter,  and that is all
	# normalization that is possible to do on the type.
	#
	# For those, it is possible to use this normalizer by doing
	# class <<self
	# 	alias_method :normalize_args, :normalize_args_sort
	# end
	#
	def Type.normalize_args_sort_uniq(args)
		normalize_arg_refs(args).sort { |a, b| a.id <=> b.id }.uniq
	end
	#
	# Normalize args to a type.  The default is to just normalize argument
	# references to be to other, equivalent objects, 
	# but this can be overridden in child classes (as done in
	# Type::Respond to use normalize_args_sort).
	#
	class <<self
		alias_method :normalize_args, :normalize_arg_refs
	end

	#
	# Set of objects of this Type, indexed based on arguments.
	# This is used as a cache to get a single copy of a type when
	# it is generated multiple times (e.g, Type::Respond(:to_s, :to_i)
	# at one point in the program will return the exact same object
	# as Type::Respond(:to_i, :to_s) another place in the program.
	# (Arguments are normalized before caching, so the different
	# order of arguments to Type::Respond does not matter.)
	#
	@@type_objs = {}
	#
	# Generate a copy of a type.
	#
	# If an equivalent type has been generated before, this will be retrieved
	# from the object cache instead.
	#
	def Type.create(*args, &block)
		# Normalize args to get a single copy for anything that
		# is equivalent
		args = normalize_args(args)
		# Generate a hash id for the above - every hash id based on
		# the same set of args will be the same, and every that
		# is different will be different.
		hashid = args.collect { |arg| arg.id }.join('|') + "|" + block.id.to_s
		unless @@type_objs[hashid]
			# Cache the arguments we got in order to avoid their object
			# ids being reused for different objects at another place in the
			# program.
			# This must be done in order to make sure the hashid
			# above is unique.
			@type_args = args.dup
			# Cache a type object that match the generated type
			@@type_objs[hashid] = new(*args) 
		end
		return @@type_objs[hashid]
	end
	# Clients are supposed to use create or Type::<subtype>(args)
	private_class_method :new

	#
	# Make it possible to create types as Type::<type>(args)
	# E.g. Type::Respond(:to_i)
	#
	def Type.method_missing(symbol, *args, &block)
		if symbol.to_s =~ /^[A-Z]/ && const_defined?(symbol) && const_get(symbol).kind_of?(Class)
			begin
				const_get(symbol).create(*args, &block)
			rescue
				raise $!, $!.message, caller
			end
		else
			raise NameError, "undefined method `#{symbol}' for #{name}:#{type}", caller
		end
	end
	#
	# Define a === by way of a call to verify (opposed to the default of
	# defining verify by way of ===.)
	# Useful for composite types like Any, All, or Multi.
	# Use by
	#	alias_method :===, :tripleequals_by_verify
	# in the class definition.
	#
	def tripleequals_by_verify(other, set_message)
		skip, msg = verify(0, -1, false, [other], self)
		return !!skip
	end
end

#
# Use of Type for pure namespacing SEPARATED from the actual type stuff.
# This may want to move away
#
class Type
	#
	# Type verifier for types defined exclusively by what messages they respond_to.
	#
	class Respond < self; extend TypeChecker
		class << self
			alias_method :normalize_args, :normalize_args_sort_uniq
		end
		add_typesig :initialize, :rest, Symbol
		def ===(other, set_message=false)
			@args.each { |symbol|
				unless other.respond_to? symbol
					set_message("does not respond to #{symbol}") if set_message
					return false
				end
			}
			return true
		end
	end

	#
	# Make a new type that use the passed block to check.
	#
	# Usage:
	#	Type::Make("Isn't numeric string") { |arg| arg =~ /^\d+$/ }
	#
	# Note:
	#	Using a block to check something is quite a bit less efficent
	#	than using the built-in types, as the built-in types takes
	#	special care to only create one instance per unique type, 
	#	are written with efficiency in mind, and require one less call-level.
	#	XXX The following is not implemented yet
	#	Also, built-in types can be compiled to code that runs as part of the
	#	actual ruby call generated to check types, instead of running through
	#	the Type interface.
	class Make < self
		def initialize(message, &block)
			raise ArgumentError, "Missing block to #{type.name}", caller \
				unless block_given?
			@message = message
			@block = block
			super
		end
		def ===(other, set_message=false)
			@block.call(other) || set_message(@message, set_message)
		end
	end

	#
	# Implementation of a simple 'or' type, allowing match of any of the
	# types.
	#
	# The skip will be the SMALLEST of the ones allowed by any of the
	# types in the list, in order to allow more specific typechecking later
	# in the typesig this is part of.
	#
	class Any < self
		alias_method :===, :tripleequals_by_verify
		def initialize(*args)
			args.each_index { |argno|
				arg = args[argno]
				if arg.kind_of? Type
				elsif arg.kind_of? Module
				else
					raise ArgumentError,
"Arg #{argno} to #{type.name} is invalid (expected [Module, Type], got #{arg.type.name}"
				end
			}
			super
		end
		def Any.normalize(*args)
			normalize_args_sort_uniq(args.flatten)
		end
		def verify(argno, arity, exact, args, *typesig)
			minskip = nil
			@args.each { |arg|
				if arg.kind_of?(Type)
					offset, string = arg.verify(argno, arity, exact, args, *typesig)
					next if string
					minskip = offset if !minskip || minskip > offset
					break if minskip == 1
				elsif arg.kind_of?(Module)
					if ((!exact && args[argno].kind_of?(arg)) ||
					    (exact && args[argno].instance_of?(arg)))
						minskip = 1
						break
					end
				else
					raise ArgumentError, "Inccorect type to #{inspect}"
				end
			}
			if minskip
				return minskip, nil
			else
				return nil, \
"is of invalid type (expected " + (exact ? "exact " : "") + "match in [#{@args.join(', ')}], got #{args[argno].type})"
			end
		end
	end
	
	#
	# Implementation of a simple 'and' type, allowing match of
	# all of a set of types.
	#
	class All < Type
		class << self
			alias_method :normalize_args, :normalize_args_sort_uniq
		end
		alias_method :===, :tripleequals_by_verify
		def verify(argno, arity, exact, args, *typesig)
			minskip = nil
			@args.each { |arg|
				failed = false
				if arg.kind_of?(Type)
					offset, string = arg.verify(argno, arity, exact, args, *typesig)
					failed = true if string
					minskip = offset if !minskip || minskip > offset
					break if minskip == 1
				elsif arg.kind_of?(Module)
					if ((!exact && args[argno].kind_of?(arg)) ||
					    (exact && args[argno].instance_of?(arg)))
						minskip = 1
						break
					else
						failed = true
					end
				else
					raise ArgumentError, "Inccorect type to #{inspect}"
				end
			}
			if minskip && !failed
				return minskip, nil
			else
				return nil, \
"is of invalid type (expected " + (exact ? "exact " : "") + "match with [#{@args.join(', ')}], got #{args[argno].type})"
			end
		end
	end

	#
	# Type that verifies N other, ordered types in a typesig.
	#
	class Multi
		def verify(argno, artiy, exact, args, *typesig)
		end
	end
end
